#!/usr/bin/env python3

import argparse
import json
import os
import re
import requests
import subprocess
import urllib3

import gitlab

from jenkinsapi.jenkins import Jenkins

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

### Well known Gitlab project IDs
GRAND_CENTRAL_PROJECT_ID = "2905"
IDENTITY_SERVICE_PROJECT_ID = "2793"
AUTH_SERVICE_PROJECT_ID = "2786"

GOV_JENKINS_URL = "https://cb-jenkins-stg.vdp-int-stg.vmware.com"

### Script expects those environment variables to be set
GITLAB_API_TOKEN = os.environ["DEVOPS_ACCESS_TOKEN"]
PROTOCOL = os.environ["CI_SERVER_PROTOCOL"]
HOST = os.environ["CI_SERVER_HOST"]
GITLAB_URL = "{}://{}".format(PROTOCOL, HOST)
#PORT = os.environ["CI_SERVER_PORT"]
PROJECT_ID = os.environ['CI_PROJECT_ID']
JENKINS_API_TOKEN = os.environ["JENKINS_API_TOKEN"]
JENKINS_USER = os.environ["JENKINS_USER"]

# Clients initialization
GITLAB_CLIENT = gitlab.Gitlab(GITLAB_URL, private_token=GITLAB_API_TOKEN, ssl_verify=False)
JENKINS_CLIENT = Jenkins(GOV_JENKINS_URL, username=JENKINS_USER, password=JENKINS_API_TOKEN)


def parse_args():
    parser = argparse.ArgumentParser(description="GSTG and GPRD deployment automation tool")
    subparsers = parser.add_subparsers(dest="command")

    
    gstg_parser = subparsers.add_parser("gstg",
                                        help="Performs a deployment to GSTG.")
    gstg_parser.add_argument("--pipeline-id",
                            dest="pipeline_id",
                            help="Pipeline ID of the pipeline that successfully deployed to GINT",
                            required=True)
    gstg_parser.add_argument("--deploy-job-name",
                            dest="deploy_job_name",
                            help="The name of the job that deploys to GINT.",
                            required=False,
                            default="GINT_DEPLOY")
    gstg_parser.add_argument("--publish-job-name",
                            dest="publish_job_name",
                            help="The name of the job that publishes Docker images and Helm charts.",
                            required=False,
                            default="build-push-artifacts")

    gprd_parser = subparsers.add_parser("gprd",
                                        help="Performs a deployment to GPRD.")
    gprd_parser.add_argument("--identity-pipeline-id",
                        dest="identity_pipeline_id",
                        help="Pipeline ID of the pipeline that successfully deployed 'platform-identity' to GSTG",
                        required=False,
                        default=None)
    gprd_parser.add_argument("--auth-pipeline-id",
                        dest="auth_pipeline_id",
                        help="Pipeline ID of the pipeline that successfully deployed 'platform-auth' to GSTG",
                        required=False,
                        default=None)
    gprd_parser.add_argument("--gstg-deploy-job-name",
                            dest="gstg_deploy_job_name",
                            help="The name of the job that deploys to GSTG.",
                            required=False,
                            default="GSTG_DEPLOY")

    args = parser.parse_args()
    return args


def jenkins_gstg_deploy(job_name, transaction_id, deploy_mode="sync"):
    if JENKINS_CLIENT.has_job(job_name):
        job = JENKINS_CLIENT.get_job(job_name)
        params = { "ARTIFACT_TRANSACTION_ID": transaction_id, "DEPLOY_MODE": deploy_mode }
        queue_item = job.invoke(build_params=params)
        print("Triggered '{}' job.".format(job_name))
        if queue_item.is_queued() or queue_item.is_running():
            print("Waiting for the build to finish...")
            queue_item.block_until_complete()
        
        build_number = queue_item.get_build_number()
        build = job.get_build(build_number)
        return build
        

def get_job(project_id, pipeline_id, job_name):
    project = GITLAB_CLIENT.projects.get(project_id)
    pipeline = project.pipelines.get(pipeline_id)
    jobs = pipeline.jobs.list()

    found_job = None
    for job in jobs:
        if job.name == job_name:
            found_job = job

    if not found_job:
        print("Job '{}' was not found in pipeline with ID '{}' for project '{}'.".format(job_name, pipeline_id, project_id))
        return None

    if found_job.status != "success":
        print("Job '{}' from pipeline with ID '{}' from project '{}' is not successfull.".format(job_name, pipeline_id, project_id))
        return None

    return found_job


def deploy_gstg(project_id, pipeline_id, deploy_job_name, publish_job_name, grand_central_project_id):
    print("Deploying to GSTG")
    print("====================================")
    deploy_job = get_job(project_id, pipeline_id, deploy_job_name)
    publish_job = get_job(project_id, pipeline_id, publish_job_name)
    
    if not deploy_job or not publish_job:
        print("Exiting.")
        exit(1)

    deploy_job_trace = get_job_trace(project_id, deploy_job.id)
    gc_pipeline_urls = [ line.split(' ')[-1] for line in deploy_job_trace if "pipeline url :" in line ]
    
    publish_job_trace = get_job_trace(project_id, publish_job.id)
    published_helm_charts = [ line.split(' ')[-1] for line in publish_job_trace if "Pushed: " in line ]
    pushed_docker_tags = [ line.split(' ')[-1] for line in publish_job_trace if "Successfully tagged " in line and ":latest" not in line ]
    
    ### Get the pipeline for 'main' branch in Grand Central and its jobs
    gc_project = GITLAB_CLIENT.projects.get(grand_central_project_id)
    gc_pipeline_main = None
    for gc_pipeline_url in gc_pipeline_urls:
        gc_pipeline_id = gc_pipeline_url.split('/')[-1]
        gc_pipeline = gc_project.pipelines.get(gc_pipeline_id)
        if gc_pipeline.ref == 'main':
            gc_pipeline_main = gc_pipeline
            break

    gc_pipeline_main_jobs = gc_pipeline_main.jobs.list()
    scan_reports_artifact = "scan_reports.tgz"
    transaction_id = ""
    bundle_job_url = ""
    for job in gc_pipeline_main_jobs:
        if 'gstg1usgw1.' in job.name and 'bundle_ship' in job.name:
            bundle_job_url = job.web_url
            trace = get_job_trace(grand_central_project_id, job.id)
            for line in trace:
                if 'TransactionId' in line:
                    transaction_id = line.split(' ')[-1]

            job_obj = gc_project.jobs.get(job.id, lazy=True)
            print("Getting the PRODSEC scan reports artifact...")
            with open(scan_reports_artifact, "wb") as f:
                try:
                    job_obj.artifact(scan_reports_artifact, streamed=True, action=f.write)
                except:
                    print("Could not download the scan reports from '{}'. Exiting.".format(job.web_url))
                    exit(1)

            subprocess.run(["tar", "-xzvf", scan_reports_artifact])

    jenkins_build = jenkins_gstg_deploy("grand-central-stg", transaction_id)
    jenkins_build_url = jenkins_build.get_build_url()

    if jenkins_build.is_good():
        print("Successfully deployed to GSTG. URL: {}".format(jenkins_build_url))
    else:
        print("GSTG deployment job failed. URL: {}".format(jenkins_build_url))
        exit(1)

    with open("output.json", "w") as output_file:
        outputs = { "docker_tags": pushed_docker_tags, 
                    "helm_charts": published_helm_charts, 
                    "transaction_id": transaction_id,
                    "bundle_job_url": bundle_job_url,
                    "jenkins_build_url": jenkins_build_url }
        json.dump(outputs, output_file, indent=4)
        print("Data saved to 'output.json'")
        print("Done.")


def strip_ansi_codes(string):
    return re.sub('\033\\[([0-9]+)(;[0-9]+)*m', '', string) 


def get_job_trace(project_id, job_id):
    headers = { "PRIVATE-TOKEN": GITLAB_API_TOKEN }

    r = requests.get(GITLAB_URL + "/api/v4/projects/{}/jobs/{}/trace".format(project_id, job_id), headers=headers, verify=False)
    
    trace = list() 
    for line in r.text.splitlines():
        trace.append(strip_ansi_codes(line))

    return trace


def get_job_obj(project_id, job_id):
    project = GITLAB_CLIENT.projects.get(project_id)
    return project.jobs.get(job_id, lazy=True)


def get_gstg_outputs(project_id, pipeline_id, gstg_deploy_job_name):
    gstg_deploy_job = get_job(project_id, pipeline_id, gstg_deploy_job_name)
    job_obj = get_job_obj(project_id, gstg_deploy_job.id)
    
    print("Getting '{}' artifacts.".format(gstg_deploy_job_name))
    artifacts_output_archive = "{}_artifacts.tgz".format(project_id)
    artifacts_dir = "{}_artifacts".format(project_id)
    with open(artifacts_output_archive, "wb") as f:
            try:
                job_obj.artifact(artifacts_output_archive, streamed=True, action=f.write)
            except:
                print("Could not download the artifacts from '{}'. Exiting.".format(gstg_deploy_job.web_url))
                exit(1)

            subprocess.run(["mkdir", "-p", artifacts_dir])
            subprocess.run(["tar", "-xzvf", artifacts_output_archive, "-C", artifacts_dir])
    
    gstg_outputs = json.load(artifacts_dir + "/output.json")
    gstg_outputs['artifacts_dir'] = artifacts_dir
    return gstg_outputs


def create_cab_ticket(identity_outputs, auth_outputs):
    print("Identity: ")
    print(identity_outputs)
    print("Auth: ")
    print(auth_outputs)
    print("Not implemented yet")


def main():
    args = parse_args()

    if args.command == "gstg":
        deploy_gstg(PROJECT_ID, args.pipeline_id, args.deploy_job_name, args.publish_job_name, GRAND_CENTRAL_PROJECT_ID)
    elif args.command == "gprd":
        identity_gstg_outputs = None
        auth_gstg_outputs = None
        if args.identity_pipeline_id:
            identity_gstg_outputs = get_gstg_outputs(IDENTITY_SERVICE_PROJECT_ID, args.identity_pipeline_id, args.gstg_deploy_job_name)
        if args.auth_pipeline_id:
            auth_gstg_outputs = get_gstg_outputs(AUTH_SERVICE_PROJECT_ID, args.auth_pipeline_id, args.gstg_deploy_job_name)

        if not args.identity_pipeline_id and not args.auth_pipeline_id:
            print("Provide a pipeline ID for at least one service via `--identity-pipeline-id` or `--auth-pipeline-id`")
            exit(1)
        else:
            create_cab_ticket(identity_gstg_outputs, auth_gstg_outputs)

    return 0


if __name__ == "__main__":
    main()
