#!/usr/bin/env python3

import argparse
import json
import os
import re
import requests
import urllib3

import gitlab

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

### Script expects those environment variables to be set
GITLAB_API_TOKEN = os.environ["DEVOPS_ACCESS_TOKEN"]
GITLAB_URL = os.environ["CI_SERVER_URL"]

# Client initialization
GITLAB_CLIENT = gitlab.Gitlab(GITLAB_URL, private_token=GITLAB_API_TOKEN, ssl_verify=False)



def parse_args():
    parser = argparse.ArgumentParser(description="Gitlab automation tool")
    subparsers = parser.add_subparsers(dest="command")


    merge_parser = subparsers.add_parser("merge", help="Performs MR actions.")
    merge_parser.add_argument("--project-id",
                            dest="project_id",
                            help="Project ID",
                            required=True)
    merge_parser.add_argument("--source-branch",
                            dest="source_branch",
                            help="Source branch name",
                            required=True)
    merge_parser.add_argument("--target-branch",
                            dest="target_branch",
                            help="Target branch name",
                            required=True,
                            default="master")
    merge_parser.add_argument("--title",
                            dest="title",
                            help="MR title",
                            required=True)
    merge_parser.add_argument("--description",
                            dest="description",
                            help="MR description",
                            required=True)
    merge_parser.add_argument("--squash",
                            dest="squash",
                            action='store_true',
                            help="Squash commits",
                            required=False,
                            default=False)
    merge_parser.add_argument("--remove-source-branch",
                            dest="remove_source_branch",
                            action='store_true',
                            help="Remove source branch",
                            required=False,
                            default=False)

    args = parser.parse_args()
    return args


def get_job(project_id, pipeline_id, job_name):
    project = GITLAB_CLIENT.projects.get(project_id)
    pipeline = project.pipelines.get(pipeline_id)
    jobs = pipeline.jobs.list()

    found_job = None
    for job in jobs:
        if job.name == job_name:
            found_job = job

    if not found_job:
        print("Job '{}' was not found in pipeline with ID '{}' for project '{}'.".format(job_name, pipeline_id, project_id))
        return None

    if found_job.status != "success":
        print("Job '{}' from pipeline with ID '{}' from project '{}' is not successfull.".format(job_name, pipeline_id, project_id))
        return None

    return found_job


def strip_ansi_codes(string):
    return re.sub('\033\\[([0-9]+)(;[0-9]+)*m', '', string) 


def get_job_trace(project_id, job_id):
    headers = { "PRIVATE-TOKEN": GITLAB_API_TOKEN }

    r = requests.get(GITLAB_URL + "/api/v4/projects/{}/jobs/{}/trace".format(project_id, job_id), headers=headers, verify=False)
    
    trace = list() 
    for line in r.text.splitlines():
        trace.append(strip_ansi_codes(line))

    return trace


def get_job_obj(project_id, job_id):
    project = GITLAB_CLIENT.projects.get(project_id)
    return project.jobs.get(job_id, lazy=True)


def create_merge_request(project_id, source_branch, target_branch, 
                         title, description, squash=True, remove_source_branch=True):
    project = GITLAB_CLIENT.projects.get(project_id, lazy=True)
    payload = { 'source_branch': source_branch,
                'target_branch': target_branch,
                'title': title,
                'description': description,
                'squash': squash,
                'remove_source_branch': remove_source_branch }
    
    mr = project.mergerequests.create(payload)
    
    return mr.web_url
        

def main():
    args = parse_args()
    
    if args.command == "merge":
        url = create_merge_request(args.project_id, args.source_branch, args.target_branch, args.title, args.description, args.squash, args.remove_source_branch)
        print("URL: {}".format(url))

    return 0


if __name__ == "__main__":
    main()
